/* driver.c - MSDOS device driver functions				*/
/*									*/
/* Copyright (C) 1994 by Robert Armstrong				*/
/*									*/
/* This program is free software; you can redistribute it and/or modify	*/
/* it under the terms of the GNU General Public License as published by	*/
/* the Free Software Foundation; either version 2 of the License, or	*/
/* (at your option) any later version.					*/
/*									*/
/* This program is distributed in the hope that it will be useful, but	*/
/* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-	*/
/* ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General	*/
/* Public License for more details.					*/
/*									*/
/* You should have received a copy of the GNU General Public License	*/
/* along with this program; if not, visit the website of the Free	*/
/* Software Foundation, Inc., www.gnu.org.				*/


#include <stdio.h>		/* NULL, etc...				*/
#include <dos.h>		/* used only for MK_FP !		*/
#include <stdarg.h>		/* needed for variable argument lists	*/
#include "standard.h"		/* definitions for this project		*/
#include "tu58.h"		/* TU58 interface and protocol		*/
#include "serial.h"		/* COM port routines (open, baud, etc)	*/
#include "driver.h"		/* MSDOS device driver interface	*/
#include "cprint.c"


#define FORM_FACTOR	8	/* DOS form factor code used for TU58s	*/
#define CACHE_SIZE	5	/* cache the FAT and the root directory	*/


/* Forward declarations for routines that need it... */
PRIVATE BOOLEAN parse_options (char far *);
PUBLIC void Initialize (rh_init_t far *);
PUBLIC void far TUDriver (rh_t far *);


/* These data structures are exported by the HEADER.ASM module... */
extern devhdr_t header;		/* MSDOS device header for our driver	*/
extern bpb_t bpb;		/* BIOS Parameter block "   "   "	*/
extern bpbtbl_t bpbtbl;		/* BPB table (one for each drive unit)	*/

/*   This double word is actually a far pointer to the entry point of	*/
/* this module. It's called by the assembly interface whenver DOS needs	*/
/* driver action.  The second word of the far pointer contains the seg-	*/
/* ment address, which is actually computed and written by the assembly	*/
/* code.								*/
PUBLIC WORD c_driver[2] = {(WORD) &TUDriver, 0};


/* Local data for this module... */
WORD    DrivePort    = 2;	/* COM port specified for the drive	*/
WORD    DriveBaud    = 9600;	/* baud rate  "   "    "   "    "	*/
BOOLEAN Debug        = FALSE;	/* TRUE to enable debug (verbose) mode	*/
BOOLEAN InitNeeded   = TRUE;	/* TRUE if we need to (re) initialize	*/
#ifdef CACHE_SIZE
BOOLEAN CacheEnabled = FALSE;	/* TRUE if the tape cache is enabled	*/
BYTE    CacheUnit[CACHE_SIZE];	/* unit number owning this cache block	*/
BYTE    TapeCache[CACHE_SIZE*BLOCKSIZE];
#endif
WORD	RebootVector[2];	/* previous INT 19H vector contents	*/


/* fmemcpy */
/*   This function is equivalent to the C RTL _fmemcpy routine. We have	*/
/* to define it here because BPC is unable to generate inline code for	*/
/* _fmemcpy (although it can generate equivalent code for memcpy!).	*/
void fmemcpy (void far *dst, void far *src, WORD n)
{
  _asm {
	les	di,dword ptr dst
	mov	dx,word ptr src+2
	mov	si,word ptr src
	mov	cx,n
	push	ds
	mov	ds,dx
	rep 	movsb
	pop	ds
  }
}

/* fmemset */
/*    This is the equivalent to _fmemset.  It is here for exactly the	*/
/* same reasons as fmemcpy !!!						*/
void fmemset (void far *dst, BYTE c, WORD n)
{
  _asm {
	mov	al,c
	les	di,dword ptr dst
	mov	cx,n
	rep 	stosb
  }
}


/* cache_sweep */
/*   This function will invalidate every block in the cache.  Since the	*/
/* cache is write thru, no actual tape I/O needs to take place.  All we	*/
/* do is to set the unit corresponding to each cache block to be 255.	*/
#ifdef CACHE_SIZE
void cache_sweep (void)
{
  if (Debug)  cprintf("TU58: cache sweep\n");
/*  for (i = 0;  i < CACHE_SIZE;  ++i)  CacheUnit[i] = 0xFF; */
  fmemset(CacheUnit, 0xFF, CACHE_SIZE);
}
#endif


/* Media Check */
/*   DOS calls this function to determine if the tape in the drive has	*/
/* been changed.  The TU58 hardware can't determine this (like many	*/
/* older 360K floppy drives), and so we always return the "Don't Know"	*/
/* response.  This works well enough...					*/
PUBLIC void MediaCheck (rh_media_check_t far *rh)
{
  if (Debug) cprintf("TU58: media check: unit=%d\n", rh->rh.unit);
  rh->media_status = 0;
  rh->rh.status = DONE;
}


/* Build BPB */
/*   DOS uses this function to build the BIOS parameter block for the	*/
/* specified drive.  For diskettes, which support different densities	*/
/* and formats, the driver actually has to read the BPB from the boot	*/
/* sector on the disk.  The TU58 is much simpler - since all tapes have	*/
/* the same format we can just use the hardwired BPB in memory...	*/
PUBLIC void BuildBPB (rh_get_bpb_t far *rh)
{
  if (Debug) cprintf("TU58: build BPB: unit=%d\n", rh->rh.unit);
  rh->bpb = &bpb;
  rh->rh.status = DONE;
#ifdef CACHE_SIZE
  cache_sweep();
#endif
}


/* Get Parameters */
/*   This routine implements the Get Parameters subfunction of the DOS	*/
/* Generic IOCTL call. It gets a pointer to the device paramters block,	*/
/* which it then fills in.  We do NOT create the track/sector map that	*/
/* is defined by recent DOS manuals to be at the end of the block - it	*/
/* never seems to be used...						*/
PUBLIC void GetParameters (device_params_t far *dp)
{
  dp->form_factor = FORM_FACTOR;  dp->attributes = 0;  dp->media_type = 0;
  dp->cylinders = bpb.total_sectors / (bpb.track_size * bpb.head_count);
  fmemcpy(&(dp->bpb), &bpb, sizeof(bpb_t));
}


/* dos_error */
/*   This routine will translate a TU58 error code into an appropriate	*/
/* DOS error code.  This driver never retries on any error condition.	*/
/* For actual tape read/write errors it's pointless because the drive	*/
/* will have already tried several times before reporting the failure.	*/
/* All the other errors (e.g. write lock, communications failures, etc)	*/
/* are not likely to succeed without user intervention, so we go thru	*/
/* the usual DOS "Abort, Retry or Ignore" dialog. Communications errors	*/
/* are a special situation.  In these cases we also set global flag to	*/
/* force a controller initialization before the next operation.		*/
int dos_error (int status)
{
  switch (status) {
    case E_SUCCESS:
    case E_WRETRY:	return 0;

    case E_WLOCK:	return WRITE_PROTECT;
    case E_BADUNIT:	return UNKNOWN_UNIT;
    case E_NOTAPE:	return NOT_READY;
    case E_DATACHECK:	return CRC_ERROR;
    case E_EOT:
    case E_SEEKFAIL:	return SEEK_ERROR;
    case E_BADBLOCK:	return BAD_SECTOR;
    case E_JAMMED:
    case E_POSTFAIL:	return GENERAL_FAILURE;

    case E_BADOPCODE:
    case E_TIMEOUT:
    case E_BADTYPE:
    case E_BADLENGTH:
    case E_CHECKSUM:	InitNeeded = TRUE;	return GENERAL_FAILURE;

    default:
	cprintf("TU58: unknown drive error - status = 0x%2x\n", status);
        return GENERAL_FAILURE;
  }
}


/* drive_init */
/*   This routine should be called before every I/O function.  If the	*/
/* last I/O operation resulted in a protocol error, then this routine	*/
/* will re-initialize the drive and the communications.  If the drive	*/
/* still won't talk to us, then it will set a general failure code in	*/
/* the request header and return FALSE.					*/
BOOLEAN drive_init (rh_t far *rh)
{
  if (!InitNeeded)  return TRUE;
  if (!TUInitialize()) {
    if (Debug)  cprintf("TU58: drive failed to initialize\n");
    rh->status = DONE | ERROR | GENERAL_FAILURE;
    return FALSE;
  }
//  InitNeeded = FALSE;
  if (Debug) cprintf("TU58: drive initialized\n");
  return TRUE;
}


/* Read Data */
PUBLIC void ReadBlock (rh_io_t far *rh)
{
  WORD lbn, count;  int status;  BYTE far *dta;
  if (Debug)
    cprintf("TU58: read block: unit=%d, start=%d, count=%d, dta=%4x:%4x\n",
      rh->rh.unit, rh->start, rh->count, FP_SEG(rh->dta), FP_OFF(rh->dta));
  if (!drive_init ((rh_t far *) rh))  return;
  count = rh->count,  lbn = rh->start,  dta = rh->dta;
  while (count > 0) {
#ifdef CACHE_SIZE
    if (    (lbn > 0)  &&  (lbn <= CACHE_SIZE)
        &&  CacheEnabled  &&  (CacheUnit[lbn-1] == rh->rh.unit) ) {
      fmemcpy(dta, TapeCache + ((lbn-1) * BLOCKSIZE), BLOCKSIZE);
      cprintf("TU58: cache hit, block=%d\n", lbn);
    } else {
#endif
      status = TURead(rh->rh.unit, lbn, dta);
      if (status != E_SUCCESS  &&  status != E_WRETRY)  {
	if (Debug) cprintf("TU58: read error - status=%d\n", status);
	fmemset(dta, 0, BLOCKSIZE);
	rh->rh.status = DONE | ERROR | dos_error(status);
	return;
      }
#ifdef CACHE_SIZE
	else if ((lbn > 0)  &&  (lbn <= CACHE_SIZE)  &&  CacheEnabled) {
	    fmemcpy(TapeCache + ((lbn-1) * BLOCKSIZE), dta, BLOCKSIZE);
	    CacheUnit[lbn-1] = rh->rh.unit;
	}
    }
#endif
    ++lbn,  --count,  dta += BLOCKSIZE;
  }
  rh->rh.status = DONE;
}


/* Write Data */
/* Write Data with Verification */
PUBLIC void WriteBlock (rh_io_t far *rh, BOOLEAN verify)
{
  WORD lbn, count;  int status;  BYTE far *dta;
  if (Debug)
    cprintf("TU58: write block: unit=%d, start=%d, count=%d, dta=%4x:%4x\n",
      rh->rh.unit, rh->start, rh->count, FP_SEG(rh->dta), FP_OFF(rh->dta));
  if (!drive_init ((rh_t far *) rh))  return;
  count = rh->count,  lbn = rh->start,  dta = rh->dta;
  while (count > 0) {
#ifdef CACHE_SIZE
    if ((lbn > 0)  &&  (lbn <= CACHE_SIZE)  &&  CacheEnabled) {
      fmemcpy(TapeCache + ((lbn-1) * BLOCKSIZE), dta, BLOCKSIZE);
      CacheUnit[lbn-1] = rh->rh.unit;
    }
#endif
    status = TUWrite(rh->rh.unit, lbn, dta, verify);
    if (status != E_SUCCESS  &&  status != E_WRETRY)  {
      if (Debug) cprintf("TU58: write error - status=%d\n", status);
      rh->rh.status = DONE | ERROR | dos_error(status);
      return;
    }
    ++lbn,  --count,  dta += BLOCKSIZE;
  }
  rh->rh.status = DONE;
}


/* Generic IOCTL */
/*   The generic IOCTL functions are used by DOS programs to determine	*/
/* the device geometry, and especially by the FORMAT program to init-	*/
/* ialize new media.  The DOS format program requires us to implement	*/
/* these three generic IOCTL functions:					*/

PUBLIC void GenericIOCTL (rh_generic_ioctl_t far *rh)
{
  if (Debug)
    cprintf("TU58: generic IOCTL: unit=%d, major=0x%2x, minor=0x%2x, data=%4x:%4x\n",
    rh->rh.unit, rh->major, rh->minor, FP_SEG(rh->packet), FP_OFF(rh->packet));

  if (rh->major == DISK_DEVICE)
    switch (rh->minor) {
      case GET_PARAMETERS:  GetParameters ((device_params_t far *) rh->packet);
			    rh->rh.status = DONE;
			    return;
      case GET_ACCESS:	    ((access_flag_t far *) (rh->packet))->allowed = 1;
			    rh->rh.status = DONE;
			    return;
      case SET_MEDIA_ID:
      case SET_ACCESS:
      case SET_PARAMETERS:
      case FORMAT_TRACK:
			    rh->rh.status = DONE;
			    return;
      case GET_MEDIA_ID:
      default: ;
    }
  cprintf("TU58: unimplemented IOCTL - unit=%d, major=0x%2x, minor=0x%2x\n",
	  rh->rh.unit, rh->major, rh->minor);
  rh->rh.status = DONE | ERROR | UNKNOWN_COMMAND;
}


/* Generic IOCTL Query */
/*   DOS programs can use this function to determine which generic	*/
/* IOCTL functions a device supports.  The query IOCTL driver call will	*/
/* succeed for any function the driver supports, and fail for others.	*/
/* Nothing actually happens - just a test for success or failure.	*/
PUBLIC void IOCTLQuery (rh_generic_ioctl_t far *rh)
{
  if (Debug)
    cprintf("TU58: generic IOCTL query: unit=%d, major=0x%2x, minor=0x%2x\n",
    rh->rh.unit, rh->major, rh->minor);
  if (rh->major == DISK_DEVICE)
    switch (rh->minor) {
      case GET_ACCESS:
      case SET_ACCESS:
      case SET_MEDIA_ID:
      case GET_PARAMETERS:
      case SET_PARAMETERS:
      case FORMAT_TRACK:	rh->rh.status = DONE;
				return;
      default:
				break;
    }
  rh->rh.status = DONE | ERROR | UNKNOWN_COMMAND;
}


/* TUDriver */
/*   This procedure is called by the DRIVER.ASM interface module when	*/
/* MSDOS calls the driver INTERRUPT routine, and the C code is expected	*/
/* to define it.  Note that the STRATEGY call is handled completely	*/
/* inside DRIVER.ASM and the address of the request header block is	*/
/* passed to the C interrupt routine as a parameter.			*/
PUBLIC void far TUDriver (rh_t far *rh)
{
/*
  if (Debug)
    cprintf("TU58: request at %4x:%4x, command=%d, length=%d\n",
      FP_SEG(rh), FP_OFF(rh), rh->command, rh->length);
*/
  switch (rh->command) {
    case INITIALIZATION:  Initialize   ((rh_init_t far *) rh);		break;
    case MEDIA_CHECK:	  MediaCheck   ((rh_media_check_t far *) rh);	break;
    case GET_BPB:	  BuildBPB     ((rh_get_bpb_t far *) rh);	break;
    case INPUT:		  ReadBlock    ((rh_io_t far *) rh);		break;
    case OUTPUT:	  WriteBlock   ((rh_io_t far *) rh, FALSE);	break;
    case OUTPUT_VERIFY:   WriteBlock   ((rh_io_t far *) rh, TRUE);	break;
    case GENERIC_IOCTL:   GenericIOCTL ((rh_generic_ioctl_t far *) rh);	break;
    case IOCTL_QUERY:     IOCTLQuery   ((rh_generic_ioctl_t far *) rh);	break;

    case GET_LOGICAL:
    case SET_LOGICAL:	  rh->status = DONE;				break;

    default:
      cprintf("TU58: unimplemented driver request - command=%d, length=%d\n",
	      rh->command, rh->length);
      rh->status = DONE | ERROR | UNKNOWN_COMMAND;
  }
}


PUBLIC void Shutdown (void)
{
  long i;
  cprintf("TU58: Shutdown\n");
  for (i=0; i <100000; ++i);
  COMClose();
  JMPVECTOR(RebootVector);
}


/*   WARNING!!  WARNING!!  WARNING!!  WARNING!!  WARNING!!  WARNING!!   */
/*									*/
/*   All code following this point in the file is discarded after the	*/
/* driver initialization.  Make absolutely sure that no routine above	*/
/* this line calls any routine below it!!				*/
/*									*/
/*   WARNING!!  WARNING!!  WARNING!!  WARNING!!  WARNING!!  WARNING!!   */


/* Driver Initialization */
/*   DOS calls this function immediately after the driver is loaded and	*/
/* expects it to perform whatever initialization is required.  Since	*/
/* this function can never be called again, it's customary to discard	*/
/* the memory allocated to this routine and any others that are used	*/
/* only at initialization.  This allows us to economize a little on the	*/
/* amount of memory used.						*/
/*									*/
/*   This routine's basic function is to initialize the serial port, go	*/
/* and make contact with the TU58, and then return a table of BPBs to	*/
/* DOS.  If we can't communicate with the drive, then the entire driver	*/
/* is unloaded from memory.						*/
PUBLIC void Initialize (rh_init_t far *rh)
{
  WORD brkadr, reboot[2];  int status, i;

  /* The version number is sneakily stored in the device header! */
  cprintf("TU58 device driver V%c.%c (C) 1994 by Robert Armstrong\n",
    header.name[6], header.name[7]);

  /* Parse the options from the CONFIG.SYS file, if any... */
  if (!parse_options((char far *) rh->bpbtbl)) {
    cprintf("TU58: bad options in CONFIG.SYS\n");
    goto unload2;
  }

  /*   Calculate the size of this driver by using the address of this	*/
  /* routine.  Note that C will return an offset for &Initialize which	*/
  /* is relative to the _TEXT segment. We have to adjust this by adding	*/
  /* the offset from DGROUP.  See HEADER.ASM for a memory layout.	*/
  brkadr = ((WORD) &Initialize) + ((_CS - _DS) << 4);
  rh->brkadr = MK_FP(_DS, brkadr);
  if (Debug)
    cprintf("TU58: CS=%4x, DS=%4x, SS=%4x, SP=%4x, break=%4x\n",
     _CS, _DS, _SS, _SP, brkadr);

  /* Initialize the COM port and set the format... */
  if (!COMOpen(DrivePort)) {
    cprintf("TU58: unable to open COM%d\n", DrivePort);
    goto unload2;
  }
  COMBaud(DriveBaud);  COMFormat("8N1");
  if (Debug)
    cprintf("TU58: initialized port COM%d at %d baud\n", DrivePort, DriveBaud);

  /* Try to make contact with the drive... */
  if (Debug) cprintf("TU58: initializing drive\n");
  if (!TUInitialize()) {
    cprintf("TU58: drive not connected or not powered\n");
    if (!Debug)  goto unload1;
  }
cprintf("TU58: rh = %4x:%4x\n", FP_SEG(rh), FP_OFF(rh));
#ifdef CACHE_SIZE
  /* Initialize the tape cache to be empty... */
  cache_sweep();
#endif
cprintf("TU58: rh = %4x:%4x\n", FP_SEG(rh), FP_OFF(rh));

  /* Execute the TU58 diagnostics and make sure it's OK... */
  if (Debug) cprintf("TU58: executing drive diagnostics\n");
  if ( (status = TUDiagnose())  !=  E_SUCCESS) {
    cprintf("TU58: drive failed diagnostics, status=%d\n", status);
    if (!Debug)  goto unload1;
  }
cprintf("TU58: rh = %4x:%4x\n", FP_SEG(rh), FP_OFF(rh));

  reboot[0] = ((WORD) &reboot) + ((_CS - _DS) << 4);
  reboot[1] = _CS;
  GETVECTOR(0x19, RebootVector);
  SETVECTOR(0x19, reboot);
  if (Debug)
    cprintf("TU58: reboot vector = %4x:%4x, old vector = %4x, %4x\n",
    reboot[1], reboot[0], RebootVector[1], RebootVector[0]);

  /* All is well.  Tell DOS how many units and the BPBs... */
  cprintf("TU58 initialized on DOS drives %c and %c\n",
    rh->drive+'A', rh->drive+'B');
  rh->nunits = 2;  rh->bpbtbl = &bpbtbl;
  rh->rh.status = DONE;
  return;

  /*   We get here if there are any errors in initialization.  In that	*/
  /* case we can unload this driver completely from memory by setting	*/
  /* (1) the break address to the starting address, (2) the number of	*/
  /* units to 0, and (3) the error flag.				*/
unload1:
  COMClose();
unload2:
  rh->brkadr = MK_FP(_DS, 0);  rh->nunits = 0;  rh->rh.status = ERROR;
}

/* iseol - return TRUE if ch is any end of line character */
PRIVATE BOOLEAN iseol (char ch)
{  return ch=='\0' || ch=='\r' || ch=='\n';  }

/* spanwhite - skip any white space characters in the string */
PRIVATE char far *spanwhite (char far *p)
{  while (*p==' ' || *p=='\t') ++p;  return p;  }

/* option_value */
/*   This routine will parse the "=nnn" part of an option.  It should	*/
/* be called with a text pointer to what we expect to be the '=' char-	*/
/* acter.  If all is well, it will return the binary value of the arg-	*/
/* ument and a pointer to the first non-numeric character.  If there is	*/
/* a syntax error, then it will return NULL.				*/
PRIVATE char far *option_value (char far *p, WORD *v)
{
  BOOLEAN null = TRUE;
  if (*p++ != '=')  return NULL;
  for (*v=0;  *p>='0' && *p<='9';  ++p)
    *v = (*v * 10) + (*p - '0'),  null = FALSE;
  return null ? NULL : p;
}

/* parse_options */
/*   This routine will parse our line from CONFIG.SYS and extract the	*/
/* driver options from it.  The routine returns TRUE if it parsed the	*/
/* line successfully, and FALSE if there are any problems.  The pointer	*/
/* to CONFIG.SYS that DOS gives us actually points at the first char-	*/
/* acter after "DEVICE=", so we have to first skip over our own file	*/
/* name by searching for a blank.  All the option values are stored in	*/
/* global variables (e.g. DrivePort, DriveBaud, etc).				*/
PRIVATE BOOLEAN parse_options (char far *p)
{
  while (*p!=' ' && *p!='\t' && !iseol(*p))  ++p;
  p = spanwhite(p);
  while (!iseol(*p)) {
    p = spanwhite(p);
    if (*p++ != '/')  return FALSE;
    switch (*p++) {
      case 'b', 'B':
        if ((p=option_value(p,&DriveBaud)) == NULL)  return FALSE;
	break;
      case 'p', 'P':
        if ((p=option_value(p,&DrivePort)) == NULL)  return FALSE;
	break;
      case 'd', 'D':
        Debug = TRUE;
	break;
#ifdef CACHE_SIZE
      case 'c', 'C':
        CacheEnabled = TRUE;
	break;
#endif
      default:
        return FALSE;
    }
    p = spanwhite(p);
  }
  return TRUE;
}
